package com.laguerre.kan;

public final class Bounds {
	private final Vector3 center = Vector3.getNewZeroVec();
	private final Vector3 size = Vector3.getNewZeroVec();
	private final Vector3 vMin = Vector3.getNewZeroVec();
	private final Vector3 vMax = Vector3.getNewZeroVec();

	public Bounds(Vector3 center, Vector3 size) {
		setValues(center, size);
	}

	public Bounds(Bounds other) {
		setValues(other.center, other.size);
	}

	public String toString() {
		return "[(" + this.center + "), size: " + this.size + "]";
	}

	public void setValues(Vector3 center, Vector3 size) {
		this.center.setValues(center);
		this.size.setValues(size);
		setMinAndMax();
	}

	public void setFromMinAndMax(Vector3 vMin, Vector3 vMax) {
		// center
		this.center.setValues(vMin);
		this.center.addSelf(vMax);
		this.center.mulSelf(0.5f);
		// size
		this.size.setValues(vMax);
		this.size.minusSelf(vMin);
		this.size.absSelf();
		// min and max
		this.vMin.setValues(vMin);
		this.vMax.setValues(vMax);
	}

	public Vector3 getSharedCenter() {
		return this.center;
	}

	public Vector3 getCenter(Vector3 result) {
		result.setValues(this.center);
		return result;
	}

	public Vector3 getCenter() {
		return new Vector3(this.center);
	}

	public Vector3 getSharedSize() {
		return this.size;
	}

	public Vector3 getSize(Vector3 result) {
		result.setValues(this.size);
		return result;
	}

	public Vector3 getSize() {
		return new Vector3(this.size);
	}

	public Vector3 getExtents(Vector3 result) {
		result.setValues(this.size);
		result.mulSelf(0.5f);
		return result;
	}

	public Vector3 getExtents() {
		final Vector3 ret = new Vector3();
		return getExtents(ret);
	}

	private void setMinAndMax() {
		final float halfSizeX = this.size.getX() / 2.0f;
		final float halfSizeY = this.size.getY() / 2.0f;
		final float halfSizeZ = this.size.getZ() / 2.0f;
		this.vMin.setValues(this.center);
		this.vMin.minusSelf(halfSizeX, halfSizeY, halfSizeZ);
		this.vMax.setValues(this.center);
		this.vMax.addSelf(halfSizeX, halfSizeY, halfSizeZ);
	}

	public void getMinAndMax(Vector3 min, Vector3 max) {
		min.setValues(this.vMin);
		max.setValues(this.vMax);
	}

	public Vector3 getSharedMin() {
		return this.vMin;
	}

	public Vector3 getSharedMax() {
		return this.vMax;
	}

	public void getHeightVertices(Vector3[] result) {
		final float halfSizeX = this.size.getX() / 2.0f;
		final float halfSizeY = this.size.getY() / 2.0f;
		final float halfSizeZ = this.size.getZ() / 2.0f;
		// Builds the 8 vertices of the bounding box.
		// xs, ys and zs can equal to -1 or +1.
		// The number 3 was chosen arbitrarily to compute -1/+1.
		int pointIndex = 0;
		for (int xs = -1; xs < 3; xs += 2) {
			for (int ys = -1; ys < 3; ys += 2) {
				for (int zs = -1; zs < 3; zs += 2) {
					result[pointIndex].setValues(
							this.center.getX() + xs * halfSizeX,
							this.center.getY() + ys * halfSizeY,
							this.center.getZ() + zs * halfSizeZ);
					pointIndex++;
				}
			}
		}
	}

	public boolean intersects(Bounds other) {
		boolean intersects = true;
		for (int i = 0; i < 3 && intersects; i++) {
			final float minVal = this.vMin.getComponent(i);
			final float maxVal = this.vMax.getComponent(i);
			final float otherMinVal = other.vMin.getComponent(i);
			final float otherMaxVal = other.vMax.getComponent(i);
			intersects =
					(minVal <= otherMinVal && otherMinVal <= maxVal) ||
					(minVal <= otherMaxVal && otherMaxVal <= maxVal);
		}
		return intersects;
	}

	public boolean contains(Vector3 point) {
		final float px = point.getX();
		final float py = point.getY();
		final float pz = point.getZ();
		final boolean notInside =
				px < vMin.getX() || py < vMin.getY() || pz < vMin.getZ() ||
				px > vMax.getX() || py > vMax.getY() || pz > vMax.getZ();
		return !notInside;
	}

	// Used by the two method 'contains' to avoid memory allocations.
	private static final Vector3 testContainPoint = Vector3.getNewZeroVec();

	public boolean contains(Bounds other) {
		boolean contained = true;
		final float halfSizeX = this.size.getX() / 2.0f;
		final float halfSizeY = this.size.getY() / 2.0f;
		final float halfSizeZ = this.size.getZ() / 2.0f;
		synchronized (testContainPoint) {
			// Builds and test the 8 vertices of the bounding box.
			// xs, ys and zs can equal to -1 or +1.
			// The number 3 was chosen arbitrarily to compute -1/+1.
			for (int xs = -1; xs < 3 && contained; xs += 2) {
				for (int ys = -1; ys < 3 && contained; ys += 2) {
					for (int zs = -1; zs < 3 && contained; zs += 2) {
						testContainPoint.setValues(
								this.center.getX() + xs * halfSizeX,
								this.center.getY() + ys * halfSizeY,
								this.center.getZ() + zs * halfSizeZ);
						contained = contains(testContainPoint);
					}
				}
			}
		}
		return contained;
	}

	// Used by the two methods 'encapsulate' to avoid memory allocations.
	private static final float[] vTmp3_0_encapsulate = new float[3];
	private static final float[] vTmp3_1_encapsulate = new float[3];
	private static final Vector3 tmpV0_encapsulate = Vector3.getNewZeroVec();
	private static final Vector3 tmpV1_encapsulate = Vector3.getNewZeroVec();

	public void encapsulate(Bounds other) {
		synchronized (vTmp3_0_encapsulate) {
			this.vMin.getComponents(vTmp3_0_encapsulate);
			this.vMax.getComponents(vTmp3_1_encapsulate);
			boolean updateNeeded = false;
			for (int i = 0; i < 3; i++) {
				if (other.vMin.getComponent(i) < this.vMin.getComponent(i)) {
					vTmp3_0_encapsulate[i] = other.vMin.getComponent(i);
					updateNeeded = true;
				}
				if (other.vMax.getComponent(i) > this.vMax.getComponent(i)) {
					vTmp3_1_encapsulate[i] = other.vMax.getComponent(i);
					updateNeeded = true;
				}
			}
			if (updateNeeded) {
				tmpV0_encapsulate.setValues(vTmp3_0_encapsulate);
				tmpV1_encapsulate.setValues(vTmp3_1_encapsulate);
				setFromMinAndMax(tmpV0_encapsulate, tmpV1_encapsulate);
			}
		}
	}

	public void encapsulate(Vector3 point) {
		synchronized (vTmp3_0_encapsulate) {
			this.vMin.getComponents(vTmp3_0_encapsulate);
			this.vMax.getComponents(vTmp3_1_encapsulate);
			boolean updateNeeded = false;
			for (int i = 0; i < 3; i++) {
				final float f = point.getComponent(i);
				if (f < vTmp3_0_encapsulate[i]) {
					vTmp3_0_encapsulate[i] = f;
					updateNeeded = true;
				}
				if (f > vTmp3_1_encapsulate[i]) {
					vTmp3_1_encapsulate[i] = f;
					updateNeeded = true;
				}
			}
			if (updateNeeded) {
				tmpV0_encapsulate.setValues(vTmp3_0_encapsulate);
				tmpV1_encapsulate.setValues(vTmp3_1_encapsulate);
				setFromMinAndMax(tmpV0_encapsulate, tmpV1_encapsulate);
			}
		}
	}
}
