package com.laguerre.kan;

import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.ShortBuffer;
import java.util.ArrayList;
import java.util.HashMap;

import android.opengl.GLES20;
import android.util.SparseArray;

public final class Shader {
	public static final class ModelRelativeData {
		public final float[] worldToLocalMatrix = new float[16];
		public final float[] localToWorldMatrix = new float[16];
		public final float[] modelMatrix = new float[16];
		public final float[] invModelMatrix = new float[16];
		public final float[] modelViewMatrix = new float[16];
		public final float[] invModelViewMatrix = new float[16];
		public final float[] modelViewProjMatrix = new float[16];
		public final float[] invModelViewProjMatrix = new float[16];

		public final void setWorldToLocalMatrix(float[] worldToLocalMatrix, float[] localToWorldMatrix) {
			System.arraycopy(worldToLocalMatrix, 0, this.worldToLocalMatrix, 0, 16);
			System.arraycopy(localToWorldMatrix, 0, this.localToWorldMatrix, 0, 16);
		}

		public final void setModelMatrix(float[] modelMatrix, float[] invModelMatrix) {
			System.arraycopy(modelMatrix, 0, this.modelMatrix, 0, 16);
			System.arraycopy(invModelMatrix, 0, this.invModelMatrix, 0, 16);
		}

		public final void setModelViewMatrix(float[] viewModelMatrix, float[] invViewModelMatrix) {
			System.arraycopy(viewModelMatrix, 0, this.modelViewMatrix, 0, 16);
			System.arraycopy(invViewModelMatrix, 0, this.invModelViewMatrix, 0, 16);
		}

		public final void setModelViewProjModelMatrix(float[] projViewModelMatrix, float[] invProjViewModelMatrix) {
			System.arraycopy(projViewModelMatrix, 0, this.modelViewProjMatrix, 0, 16);
			System.arraycopy(invProjViewModelMatrix, 0, this.invModelViewProjMatrix, 0, 16);
		}

		public final float[] getSharedWorldToLocalMatrix() {
			return this.worldToLocalMatrix;
		}

		public final float[] getSharedLocalToWorldMatrixMatrix() {
			return this.localToWorldMatrix;
		}

		public final float[] getSharedModelMatrix() {
			return this.modelMatrix;
		}

		public final float[] getSharedInvModelMatrix() {
			return this.invModelMatrix;
		}

		public final float[] getSharedModelViewMatrix() {
			return this.modelViewMatrix;
		}

		public final float[] getSharedInvModelViewMatrix() {
			return this.invModelViewMatrix;
		}

		public final float[] getSharedModelViewProjMatrix() {
			return this.modelViewProjMatrix;
		}

		public final float[] getSharedInvModelViewProjMatri() {
			return this.invModelViewProjMatrix;
		}
	}

	public static class EmptyPassShaderException extends RuntimeException {
		private static final long serialVersionUID = -5395715101387416499L;

		public EmptyPassShaderException(String msg) {
			super(msg);
		}
	}

	public static abstract class ShaderPrm {
		private static int availableId = Integer.MIN_VALUE;

		private static int grabAvailableId() {
			int id = availableId;
			++availableId;
			return id;
		}

		private int id;
		private String variableName;

		protected ShaderPrm(String variableName) {
			this.id = grabAvailableId();
			this.variableName = variableName;
		}

		public final int getId() {
			return this.id;
		}

		public final String getVariableName() {
			return this.variableName;
		}

		public abstract void passToShader(PassShader passShader);
	}

	private static class ShaderPrmInt extends ShaderPrm {
		private int data;

		protected ShaderPrmInt(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(String variableName, int n) {
			final ShaderPrmInt ret = new ShaderPrmInt(variableName);
			ret.data = n;
			return ret;
		}

		public final void updateValues(int n) {
			this.data = n;
		}

		@Override
		public void passToShader(PassShader passShader) {
			passShader.passUniformInt(getVariableName(), this.data);
		}
	}

	private static class ShaderPrmFloat extends ShaderPrm {
		private float data;

		protected ShaderPrmFloat(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(String variableName, float f) {
			final ShaderPrmFloat ret = new ShaderPrmFloat(variableName);
			ret.data = f;
			return ret;
		}

		public final void updateValues(float f) {
			this.data = f;
		}

		@Override
		public void passToShader(PassShader passShader) {
			passShader.passUniformFloat(getVariableName(), this.data);
		}
	}

	private static class ShaderPrmIntVector extends ShaderPrm {
		private int[] refData = null;

		protected ShaderPrmIntVector(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(String variableName, int[] refT) {
			final ShaderPrmIntVector ret = new ShaderPrmIntVector(variableName);
			ret.refData = refT;
			return ret;
		}

		public final void updateValues(int[] refT) {
			this.refData = refT;
		}

		@Override
		public void passToShader(PassShader passShader) {
			passShader.passUniformVec(getVariableName(), this.refData, this.refData.length);
		}
	}

	private static class ShaderPrmFloatVector extends ShaderPrm {
		private float[] refData = null;
		private boolean createdByVector3 = false;
		private boolean createdByColor= false;

		protected ShaderPrmFloatVector(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(String variableName, Vector3 v) {
			final ShaderPrmFloatVector ret = new ShaderPrmFloatVector(variableName);
			ret.refData = new float[] { v.getX(), v.getY(), v.getZ() };
			ret.createdByVector3 = true;
			return ret;
		}

		public static ShaderPrm createData(String variableName, Color c) {
			final ShaderPrmFloatVector ret = new ShaderPrmFloatVector(variableName);
			ret.refData = c.getSharedComponents();
			ret.createdByColor = true;
			return ret;
		}

		public static ShaderPrm createData(String variableName, float[] refT) {
			final ShaderPrmFloatVector ret = new ShaderPrmFloatVector(variableName);
			ret.refData = refT;
			return ret;
		}

		public final void updateValues(Vector3 v) {
			if (!this.createdByVector3) {
				throw new RuntimeException(
						"Cannot be updated by Vector3 because not created by Vector3");
			}
			this.refData[0] = v.getX();
			this.refData[1] = v.getY();
			this.refData[2] = v.getZ();
		}

		public final void updateValues(Color refC) {
			if (!this.createdByColor) {
				throw new RuntimeException(
						"Cannot be updated by Color because not created by Color");
			}
			this.refData = refC.getSharedComponents();
		}

		public final void updateValues(float[] refT) {
			if (this.createdByVector3 || this.createdByColor) {
				throw new RuntimeException(
						"Cannot be updated by float[] because not created by float[]");
			}
			this.refData = refT;
		}

		@Override
		public void passToShader(PassShader passShader) {
			passShader.passUniformVec(getVariableName(), this.refData, this.refData.length);
		}
	}

	private static class ShaderPrmMatrix extends ShaderPrm {
		private float[] refData = null;
		private int matrixType = -1;

		protected ShaderPrmMatrix(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(String variableName, float[] refM, int matrixType) {
			final ShaderPrmMatrix ret = new ShaderPrmMatrix(variableName);
			ret.refData = refM;
			ret.matrixType = matrixType;
			return ret;
		}

		public final void updateValues(float[] refM) {
			this.refData = refM;
		}

		@Override
		public void passToShader(PassShader passShader) {
			passShader.passUniformMatrix(getVariableName(), this.refData, this.matrixType);
		}
	}

	private static interface IShaderBuffer {
		public void release();
	}

	private static class ShaderPrmBuffer extends ShaderPrm implements IShaderBuffer {
		private int size = -1;
		private int stride = -1;
		private Buffer refBuffer = null;
		private int bufferPosition = -1;
		private int shaderHandle = -1;

		protected ShaderPrmBuffer(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(
				String variableName, int size, int stride, Buffer refBuffer, int bufferPosition) {
			final ShaderPrmBuffer ret = new ShaderPrmBuffer(variableName);
			ret.updateValues(size, stride, refBuffer, bufferPosition);
			return ret;
		}

		public final void updateValues(int size, int stride, Buffer refBuffer, int bufferPosition) {
			this.size = size;
			this.stride = stride;
			this.refBuffer = refBuffer;
			this.bufferPosition = bufferPosition;
		}

		@Override
		public void passToShader(PassShader passShader) {
			this.shaderHandle = GLES20.glGetAttribLocation(
	    			passShader.getProgramHandle(), getVariableName());
			if (this.shaderHandle != -1) {
				int glType;
				if (refBuffer instanceof ByteBuffer) {
					glType = GLES20.GL_BYTE;
				} else if (refBuffer instanceof ShortBuffer) {
					glType = GLES20.GL_SHORT;
				} else if (refBuffer instanceof FloatBuffer) {
					glType = GLES20.GL_FLOAT;
				} else if (refBuffer instanceof IntBuffer) {
					glType = GLES20.GL_INT;
				} else {
					throw new UnsupportedBufferTypeException(
							"Type unsupported for variable \"" +
							getVariableName() + "\"");
				}

				final int backupPosition = this.refBuffer.position();
				this.refBuffer.position(this.bufferPosition);
	    		GLES20.glEnableVertexAttribArray(this.shaderHandle);
	    		GLES20.glVertexAttribPointer(
	    				this.shaderHandle, this.size,
	    				glType, false, this.stride, this.refBuffer);
	    		// Restores the original position.
	    		this.refBuffer.position(backupPosition);
			}
		}

		@Override
		public void release() {
			if (this.shaderHandle != -1) {
				GLES20.glDisableVertexAttribArray(this.shaderHandle);
				this.shaderHandle = -1;
			}
		}
	}

	public static enum VBOPerfHint {
		STATIC_DRAW,
		STREAM_DRAW,
		DYNAMIC_DRAW
	}

	private static final class ShaderPrmVBO extends ShaderPrm {
		private static class VBOData {
			public String variableName;
			public int size;
			public int stride;
			public int startPosition;

			public VBOData(String variableName, int size, int stride, int startPosition) {
				this.variableName = variableName;
				this.size = size;
				this.stride = stride;
				this.startPosition = startPosition;
			}
		}

		private final ArrayList<VBOData> storedVBOData = new ArrayList<VBOData>();
		private int vboId = -1;
		private final VBOPerfHint perfHint;
		private boolean dataUploadedOnce = false;
		private int bufferSize = -1;
		private Buffer refBuffer = null;

		private boolean valuesUpdated = false;
		private int updatedBufferSize = -1;

		protected ShaderPrmVBO(
				String variableName,
				int bufferSize, int stride, Buffer refBuffer, int bufferPosition,
				VBOPerfHint perfHint) {
			this(variableName, bufferSize, refBuffer, perfHint);
			createVBOId();
			identifyDataInPack(variableName, bufferSize, stride, bufferPosition);
		}

		protected ShaderPrmVBO(
				String packedVBOName,
				int bufferSize, Buffer refBuffer, VBOPerfHint perfHint) {
			super(packedVBOName);
			createVBOId();
			this.bufferSize = bufferSize;
			this.updatedBufferSize = bufferSize;
			this.refBuffer = refBuffer;
			this.perfHint = perfHint;
		}

		protected final void createVBOId() {
			final int[] vboIds = new int[1];
			GLES20.glGenBuffers(1, vboIds, 0);
			this.vboId = vboIds[0];
		}

		public static ShaderPrm createData(
				String variableName,
				int size, int stride, Buffer refBuffer,
				int bufferPosition, VBOPerfHint perfHint) {
			final ShaderPrmVBO ret = new ShaderPrmVBO(
					variableName, size, stride,
					refBuffer, bufferPosition, perfHint);
			return ret;
		}

		public static ShaderPrm createPackedData(
				String packedVBOName, int globalSize, Buffer refBuffer,
				VBOPerfHint perfHint) {
			final ShaderPrmVBO ret = new ShaderPrmVBO(
					packedVBOName, globalSize, refBuffer, perfHint);
			return ret;
		}

		public final void identifyDataInPack(String variableName, int size, int stride, int startPosition) {
			final VBOData vboData = new VBOData(variableName, size, stride, startPosition);
			addVBOData(vboData);
		}

		protected boolean addVBOData(VBOData vboData) {
			boolean alreadyExists = false;
			for (int i = 0, iEnd = this.storedVBOData.size(); i < iEnd && !alreadyExists; i++) {
				if (this.storedVBOData.get(i).variableName.equals(vboData.variableName)) {
					alreadyExists = true;
				}
			}
			if (!alreadyExists) {
				this.storedVBOData.add(vboData);
			}
			return !alreadyExists;
		}

		public final void updateValues() {
			this.valuesUpdated = true;
		}

		public final void updateValues(int size, Buffer refBuffer) {
			this.updatedBufferSize = size;
			this.refBuffer = refBuffer;
			this.valuesUpdated = true;
		}

		@Override
		public void passToShader(PassShader passShader) {
			if (!this.dataUploadedOnce) {
				GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, this.vboId);
				int drawMode;
				switch (this.perfHint) {
				case STREAM_DRAW:
					drawMode = GLES20.GL_STREAM_DRAW;
					break;
				case DYNAMIC_DRAW:
					drawMode = GLES20.GL_DYNAMIC_DRAW;
					break;
				case STATIC_DRAW:
				default:
					drawMode = GLES20.GL_STATIC_DRAW;
					break;
				}
				final int backupPosition = this.refBuffer.position();
				this.refBuffer.position(0);
				GLES20.glBufferData(GLES20.GL_ARRAY_BUFFER, this.bufferSize, this.refBuffer, drawMode);
				this.refBuffer.position(backupPosition);
				GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, 0);
				this.dataUploadedOnce = true;
			} else if (this.valuesUpdated) {
				// Updates data only if we are using the stream or dynamic mode.
				switch (this.perfHint) {
				case STREAM_DRAW:
				case DYNAMIC_DRAW:
					final int backupPosition = this.refBuffer.position();
					this.refBuffer.position(0);

					GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, this.vboId);
					if (this.updatedBufferSize != this.bufferSize) {
						// The buffer must be resized: it is so replaced by a new one.
						GLES20.glBufferData(GLES20.GL_ARRAY_BUFFER,
								this.updatedBufferSize, this.refBuffer,
								this.perfHint == VBOPerfHint.DYNAMIC_DRAW ? GLES20.GL_DYNAMIC_DRAW : GLES20.GL_STREAM_DRAW);
						this.bufferSize = this.updatedBufferSize;
					} else {
						GLES20.glBufferSubData(GLES20.GL_ARRAY_BUFFER, 0, this.bufferSize, this.refBuffer);
					}
					GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, 0);
					this.refBuffer.position(backupPosition);
					break;
				default:
					break;
				}
				this.valuesUpdated = false;
			}

			for (int i = 0, iEnd = this.storedVBOData.size(); i < iEnd; i++) {
				final VBOData vboData = this.storedVBOData.get(i);
				GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, this.vboId);
				final int attribHandle = GLES20.glGetAttribLocation(
						passShader.getProgramHandle(), vboData.variableName);
				if (attribHandle != -1) {
					int glType;
					int toByteSize;
					if (refBuffer instanceof ByteBuffer) {
						glType = GLES20.GL_BYTE;
						toByteSize = Byte.SIZE / 8;
					} else if (refBuffer instanceof ShortBuffer) {
						glType = GLES20.GL_SHORT;
						toByteSize = Short.SIZE / 8;
					} else if (refBuffer instanceof FloatBuffer) {
						glType = GLES20.GL_FLOAT;
						toByteSize = Float.SIZE / 8;
					} else if (refBuffer instanceof IntBuffer) {
						glType = GLES20.GL_INT;
						toByteSize = Integer.SIZE / 8;
					} else {
						throw new UnsupportedBufferTypeException(
								"Type unsupported for variable \"" +
								vboData.variableName + "\"");
					}

					final int backupPosition = this.refBuffer.position();
					this.refBuffer.position(vboData.startPosition);

					GLES20.glEnableVertexAttribArray(attribHandle);
					AndroidGL20Compact.getInstance().glVertexAttribPointer(
							attribHandle, vboData.size,
							glType, false, vboData.stride,
							vboData.startPosition * toByteSize);
//					GLES20.glVertexAttribPointer(
//							attribHandle, vboData.size,
//							glType, false, vboData.stride,
//							vboData.startPosition * toByteSize);
					// Restores the original position.
					this.refBuffer.position(backupPosition);
				}
			}
			GLES20.glBindBuffer(GLES20.GL_ARRAY_BUFFER, 0);
		}
	}

	public static class UnsupportedBufferTypeException extends RuntimeException {
		private static final long serialVersionUID = 4692975864534105028L;

		public UnsupportedBufferTypeException(String msg) {
			super(msg);
		}
	}

	private static class ShaderPrmTexture extends ShaderPrm {
		private Texture textureRef = null;
		private int textureHandle;
		private boolean createdByTexRef = false;

		protected ShaderPrmTexture(String variableName) {
			super(variableName);
		}

		public static ShaderPrm createData(String variableName, Texture textureRef) {
			final ShaderPrmTexture ret = new ShaderPrmTexture(variableName);
			ret.textureRef = textureRef;
			ret.createdByTexRef = true;
			return ret;
		}

		public static ShaderPrm createData(String variableName, int textureHandle) {
			final ShaderPrmTexture ret = new ShaderPrmTexture(variableName);
			ret.textureHandle = textureHandle;
			ret.createdByTexRef = false;
			return ret;
		}

		public final void updateValues(Texture textureRef) {
			if (!this.createdByTexRef) {
				throw new RuntimeException(
						"Cannot be updated by textureRef because created by handle");
			}
			this.textureRef = textureRef;
		}

		public final void updateValues(int textureHandle) {
			if (this.createdByTexRef) {
				throw new RuntimeException(
						"Cannot be updated by handle because created by textureRef");
			}
			this.textureHandle = textureHandle;
		}

		@Override
		public void passToShader(PassShader passShader) {
			final int textureNum = Texture.getThenIncrTextureCounter();
			if (this.createdByTexRef) {
				passShader.passUniformTexture(getVariableName(), textureNum, this.textureRef);
			} else {
				passShader.passUniformTextureHandle(getVariableName(), textureNum, this.textureHandle);
			}
		}
	}

	public static enum RenderQueueType {
		BACKGROUND (1000),
		GEOMETRY (2000),
		TRANSPARENT (3000);

		private final int index;

		RenderQueueType(int index) {
	        this.index = index;
	    }

	    public int index() {
	        return index;
	    }
	}

	private int renderQueueType = RenderQueueType.GEOMETRY.index();
	private final ArrayList<PassShader> passShaders = new ArrayList<PassShader>();
	private int currentPassShader = 0;
	private final HashMap<String, ShaderPrm> shaderPrms = new HashMap<String, ShaderPrm>();
	/**
	 * Iterating <code>shaderPrms</code> would lead to the allocation of an
	 * Iterator. It is preferred to use this SparseArray for this reason.
	 */
	private final SparseArray<ShaderPrm> shaderPrmsIterate = new SparseArray<Shader.ShaderPrm>();
	/**
	 * Like <code>shaderPrmsIterate</code>, avoids any memory allocation from an
	 * Iterator and allows us to retrieve buffers that need to be released.
	 */
	private final SparseArray<IShaderBuffer> shaderBuffersIterate = new SparseArray<Shader.IShaderBuffer>();

	private ShaderPrm getShaderPrm(String variableName) {
		return this.shaderPrms.get(variableName);
	}

	private void addShaderPrm(ShaderPrm prm) {
		if (this.shaderPrms.put(prm.getVariableName(), prm) == null) {
			shaderPrmsIterate.put(prm.getId(), prm);
			if (prm instanceof IShaderBuffer) {
				this.shaderBuffersIterate.put(prm.getId(), (IShaderBuffer) prm);
			}
		}
	}

	private ShaderPrm removeShaderPrm(String variableName) {
		final ShaderPrm ret = this.shaderPrms.remove(variableName);
		if (ret != null) {
			this.shaderPrmsIterate.remove(ret.getId());
			if (ret instanceof IShaderBuffer) {
				this.shaderBuffersIterate.remove(ret.getId());
				((IShaderBuffer) ret).release();
			}
		}
		return ret;
	}

	private void removeAllShaderPrms() {
		this.shaderPrms.clear();
		this.shaderPrmsIterate.clear();

		for (int i = 0, iEnd = this.shaderBuffersIterate.size(); i < iEnd; i++) {
			final IShaderBuffer sbuffer = this.shaderBuffersIterate.valueAt(i);
			sbuffer.release();
		}
		this.shaderBuffersIterate.clear();
	}

	public void setRenderQueueType(RenderQueueType renderQueueType) {
		this.renderQueueType = renderQueueType.index();
	}

	public void setRenderQueueType(int renderQueueType) {
		this.renderQueueType = renderQueueType;
	}

	public int getRenderQueueType() {
		return this.renderQueueType;
	}

	public final int passCount() {
		return passShaders.size();
	}

	public final PassShader addPassShader(int passNum) {
		final int nbPasses = passCount();
		if (passNum < 0 || passNum > nbPasses) {
			throw new EmptyPassShaderException(
					"Tried to use pass " + passNum + " but passCount() = " + nbPasses);
		}
		final PassShader passShader = new PassShader();
		// Change a previous pass shader or add a new one at the end.
		if (passNum < nbPasses) {
			this.passShaders.set(passNum, passShader);
		} else {
			this.passShaders.add(passShader);
		}
		return passShader;
	}

	public final void removePassShader(int passNum) {
		this.passShaders.remove(passNum);
	}

	public final boolean setPass(int passNum) {
		this.currentPassShader = passNum;
		return isPassAvailable(passNum);
	}

	public final int getCurrentPass() {
		return this.currentPassShader;
	}

	public final boolean isPassAvailable(int passNum) {
		return passNum >= 0 && passNum < passCount();
	}

	public final void setShaderValue(String variableName, boolean b) {
		setShaderValue(variableName, b ? 1 : 0);
	}

	public final void setShaderValue(String variableName, int n) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmInt.createData(variableName, n);
			addShaderPrm(prm);
		} else {
			((ShaderPrmInt) prm).updateValues(n);
		}
	}

	public final void setShaderValue(String variableName, float f) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmFloat.createData(variableName, f);
			addShaderPrm(prm);
		} else {
			((ShaderPrmFloat) prm).updateValues(f);
		}
	}

	public final void setShaderVector(String variableName, int[] refT) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmIntVector.createData(variableName, refT);
			addShaderPrm(prm);
		} else {
			((ShaderPrmIntVector) prm).updateValues(refT);
		}
	}

	public final void setShaderVector(String variableName, Vector3 v) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmFloatVector.createData(variableName, v);
			addShaderPrm(prm);
		} else {
			((ShaderPrmFloatVector) prm).updateValues(v);
		}
	}

	public final void setShaderVector(String variableName, Color c) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmFloatVector.createData(variableName, c);
			addShaderPrm(prm);
		} else {
			((ShaderPrmFloatVector) prm).updateValues(c);
		}
	}

	public final void setShaderVector(String variableName, float[] refT) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmFloatVector.createData(variableName, refT);
			addShaderPrm(prm);
		} else {
			((ShaderPrmFloatVector) prm).updateValues(refT);
		}
	}

	public final void setShaderMatrix2(String variableName, float[] refM) {
		setShaderMatrix(variableName, refM, 2);
	}

	public final void setShaderMatrix3(String variableName, float[] refM) {
		setShaderMatrix(variableName, refM, 3);
	}

	public final void setShaderMatrix4(String variableName, float[] refM) {
		setShaderMatrix(variableName, refM, 4);
	}

	private final void setShaderMatrix(String variableName, float[] refM, int matrixType) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmMatrix.createData(variableName, refM, matrixType);
			addShaderPrm(prm);
		} else {
			((ShaderPrmMatrix) prm).updateValues(refM);
		}
	}

	public final void setShaderBuffer(
			String variableName, int size, int stride, FloatBuffer refBuffer) {
		setShaderBuffer(variableName, size, stride, refBuffer, 0);
	}

	public final void setShaderBuffer(
			String variableName, int size, int stride, FloatBuffer refBuffer, int bufferPosition) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmBuffer.createData(variableName, size, stride, refBuffer, bufferPosition);
			addShaderPrm(prm);
		} else {
			((ShaderPrmBuffer) prm).updateValues(size, stride, refBuffer, bufferPosition);
		}
	}

	public final void releaseShaderBuffer(String variableName) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm != null && prm instanceof IShaderBuffer) {
			((IShaderBuffer) prm).release();
		}
	}

	public final void releaseAllShaderBuffers() {
		for (int i = 0, iEnd = this.shaderBuffersIterate.size(); i < iEnd; i++) {
			final IShaderBuffer sbuffer = this.shaderBuffersIterate.valueAt(i);
			sbuffer.release();
		}
	}

	public final void createShaderVBO(
			String variableName, int size, int stride, FloatBuffer refBuffer,
			int bufferPosition, VBOPerfHint perfHint) {
			final ShaderPrm prm = ShaderPrmVBO.createData(
					variableName, size, stride, refBuffer, bufferPosition, perfHint);
			addShaderPrm(prm);
	}

	public final void createShaderPackedVBO(
			String packedVBOName,
			int bufferSize, FloatBuffer refBuffer, VBOPerfHint perfHint) {
			final ShaderPrm prm = ShaderPrmVBO.createPackedData(
					packedVBOName, bufferSize, refBuffer, perfHint);
			addShaderPrm(prm);
	}

	public final void identifyDataInPack(String variableOrPackedVBOName,
			String variableName, int bufferSize, int stride, int bufferPosition) {
		final ShaderPrmVBO prm = (ShaderPrmVBO) getShaderPrm(variableOrPackedVBOName);
		prm.identifyDataInPack(variableName, bufferSize, stride, bufferPosition);
	}

	public final void updateShaderVBO(
			String variableOrPackedVBOName, int size, FloatBuffer refBuffer) {
		final ShaderPrmVBO prm = (ShaderPrmVBO) getShaderPrm(variableOrPackedVBOName);
		prm.updateValues(size, refBuffer);
	}

	public final void updateShaderVBO(String variableOrPackedVBOName) {
		final ShaderPrmVBO prm = (ShaderPrmVBO) getShaderPrm(variableOrPackedVBOName);
		prm.updateValues();
	}

	public final void setShaderTexture(String variableName, Texture refTexture) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmTexture.createData(variableName, refTexture);
			addShaderPrm(prm);
		} else {
			((ShaderPrmTexture) prm).updateValues(refTexture);
		}
	}

	public final void setShaderTexture(String variableName, int textureHandle) {
		ShaderPrm prm = getShaderPrm(variableName);
		if (prm == null) {
			prm = ShaderPrmTexture.createData(variableName, textureHandle);
			addShaderPrm(prm);
		} else {
			((ShaderPrmTexture) prm).updateValues(textureHandle);
		}
	}

	public final boolean shaderVariableExists(String variableName) {
		return getShaderPrm(variableName) != null;
	}

	public final void removeShaderVariable(String variableName) {
		removeShaderPrm(variableName);
	}

	public final void removeAllShaderVariables() {
		removeAllShaderPrms();
	}

	public void use() {
		if (isPassAvailable(currentPassShader)) {
			final PassShader passShader = this.passShaders.get(this.currentPassShader);
			// Sets all the OpenGL params.
			passShader.setupRendering();
			// Sets the program to use.
			GLES20.glUseProgram(passShader.getProgramHandle());
			// Provide all the required parameters to use the current pass shader.
			for (int i = 0, iEnd = this.shaderPrmsIterate.size(); i < iEnd; i++) {
				final ShaderPrm prm = this.shaderPrmsIterate.valueAt(i);
				prm.passToShader(passShader);
			}
		}
	}

	public void destroyResources() {
		for (int i = 0, iEnd = this.passShaders.size(); i < iEnd; i++) {
			this.passShaders.get(i).destroyResources();
		}
		this.passShaders.clear();
	}
}
