package com.laguerre.kan;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import android.content.Context;
import android.opengl.GLES20;
import android.util.Log;

public final class Utils {
	/**
	 * Represents a bind to a framebuffer that has not been created.
	 */
	public static class WrongFrameBufferHandleException extends RuntimeException {
		private static final long serialVersionUID = 9066562408218138881L;

		public WrongFrameBufferHandleException(String msg) {
			super(msg);
		}
	}

	public static class FrameBufferIncompleteException extends RuntimeException {
		private static final long serialVersionUID = -8348255784559121128L;

		public FrameBufferIncompleteException(String msg) {
			super(msg);
		}
	}

	private static Context currentContext;

	public static void setContext(Context context) {
		currentContext = context;
	}

	public static Context getContext() {
		return currentContext;
	}

	public static InputStream getInputStreamFromRawResource(final int resourceId) {
		return getInputStreamFromRawResource(currentContext, resourceId);
	}

	public static InputStream getInputStreamFromRawResource(final Context context,
			final int resourceId) {
		return context.getResources().openRawResource(resourceId);
	}

	public static String readTextFileFromRawResource(final int resourceId) {
		return readTextFileFromRawResource(currentContext, resourceId);
	}

	public static String readTextFileFromRawResource(final Context context,
			final int resourceId) {
		final InputStream inputStream = context.getResources().openRawResource(
				resourceId);
		return readTextFileFromRawResource(inputStream);
	}

	public static String readTextFileFromRawResource(final InputStream inputStream) {
		final InputStreamReader inputStreamReader = new InputStreamReader(
				inputStream);
		final BufferedReader bufferedReader = new BufferedReader(inputStreamReader, 8192);

		final StringBuilder body = new StringBuilder();
		try {
			String nextLine = null;
			while ((nextLine = bufferedReader.readLine()) != null) {
				body.append(nextLine);
				body.append('\n');
			}
		} catch (IOException e) {
			return null;
		}
		return body.toString();
	}

	public static void checkGlError(String glOperation) {
		checkGlError("kan", glOperation);
	}

    /**
     * Utility method for debugging OpenGL calls. Provide the name of the call
     * just after making it:
     *
     * <pre>
     * mColorHandle = GLES20.glGetUniformLocation(mProgram, "vColor");
     * MyGLRenderer.checkGlError("glGetUniformLocation");</pre>
     *
     * If the operation is not successful, the check throws an error.
     *
     * @param glOperation - Name of the OpenGL call to check.
     * @param tag - A tag to recognize logs.
     */
    public static void checkGlError(String tag, String glOperation) {
        int error;
        while ((error = GLES20.glGetError()) != GLES20.GL_NO_ERROR) {
            Log.e(tag, glOperation + ": glError " + error);
            throw new RuntimeException(glOperation + ": glError " + error);
        }
    }

    public static float clamp(float value, float min, float max) {
    	return (value < min ? min : (max < value ? max : value));
    }

    public static int clamp(int value, int min, int max) {
    	return (value < min ? min : (max < value ? max : value));
    }

    /**
     * Interpolates <code>start</code>, at <code>t=0</code>, and
     * <code>end</code>, at <code>t=1</code>.
     * @param start
     * @param end
     * @param t between 0 and 1.
     * @return a value between, or equal to, <code>start</code> and <code>end</code>.
     */
	public static float mix(float start, float end, float t) {
		return start * (1.0f - t) + end * t;
	}

	public static float mix(float start, float end, float tMin, float tMax, float t) {
		final float t01 = (t - tMin) / (tMax - tMin);
		return mix(start, end, t01);
	}

	/**
	 * Gets a {@link Vector3} with random coordinates from/inside a sphere of radius
	 * <code>maxRadius</code>.
	 * @param result
	 * @param maxRadius
	 * @param onSurface
	 * @return
	 */
	public static Vector3 randomPointFromSphere(Vector3 result, float maxRadius, boolean onSurface) {
		final float theta = Utils.mix(0.0f, 2f * (float) Math.PI, (float) Math.random());
		final float phi = Utils.mix(0.0f, (float) Math.PI, (float) Math.random());
		final float r = onSurface ? maxRadius : Utils.mix(0.0f, maxRadius, (float) Math.random());

		result.setValues(
				(float) (r * Math.cos(theta) * Math.sin(phi)),
				(float) (r * Math.sin(theta) * Math.sin(phi)),
				(float) (r * Math.cos(phi)));
		return result;
	}

	/**
	 * Gets a {@link Vector3} with random coordinates from/inside a cone of radius
	 * <code>maxRadius</code>, height of <code>1</code> at maximum, pointing to
	 * <code>-Z</code> along axe Z.
	 * @param result
	 * @param minTheta
	 * @param maxTheta
	 * @param maxRadius
	 * @param onSurface
	 * @return
	 */
	public static Vector3 randomPointFromCone(Vector3 result,
			float maxHeight, float maxRadius, boolean onSurface) {
		final float k = Utils.mix(0.0f, 1.0f, (float) Math.random());
		final float theta = Utils.mix((float) (-Math.PI), (float) Math.PI, (float) Math.random());
		final float r = onSurface ? maxRadius : Utils.mix(0.0f, maxRadius, (float) Math.random());

		result.setValues(
				(float) (-k * r * Math.cos(theta)),
				(float) (-k * r * Math.sin(theta)),
				-k * maxHeight);
		return result;
	}

	public static boolean fuzzyEquals(float f0, float f1) {
		final float EPSILON = 1e-3f;
		return Math.abs(f0 - f1) < EPSILON;
	}
}
